/**
 * Copyright (c) 2020 xxx Inc.
 * File              : ffmpeg-backend.cc
 * Author            : 
 * Date              : 2020-05-07
 * Last Modified Date: 2020-05-07
 * Last Modified By  : 
 */
#include <vf/io/decoder-backend.h>

#if USE_FFmpeg

#include <cxxutil/logging.h>

using namespace cxxutil;
using namespace std;

namespace vf {
namespace io {

class FFmpegBackend : public DecoderBackend {
 public:
  using DecoderBackend::DecoderBackend;
  virtual ~FFmpegBackend() { Close(); }

 public:
  int Init(AVStream *video_stream, AVCodec *codec, PixelFormat tgt_pix_fmt,
           int tgt_w, int tgt_h, int sample_rate = 1,
           int64_t timestamp_intervals = -1, bool with_mvs = false) override;
  void Close() override;
  int PutPacket(const AVPacket *packet, int64_t timestamp) override;
  int Next(VFFrame &vf_frame) override;

 private:
  int PostProcessFrame(VFFrame &vf_frame);
  int InitFilters(int src_pix_fmt);

 public:
  static vector<BackendInfo> GetBackendInfo();
  static void ReleaseBackendInfo() {}
  static const string kBackendName;

 private:
  AVCodecContext *codec_ctx_ = nullptr;
  AVFrame *av_frame_ = nullptr;
  AVFrame *decoded_frame_ = nullptr;
  AVFilterGraph *filter_graph_ = nullptr;
  AVFilterContext *buffersrc_ctx_ = nullptr;
  AVFilterContext *buffersink_ctx_ = nullptr;
  AVPixelFormat av_pix_fmt_ = AV_PIX_FMT_NONE;
};

int FFmpegBackend::Init(AVStream *video_stream, AVCodec *codec,
                        PixelFormat tgt_pix_fmt, int tgt_w, int tgt_h,
                        int sample_rate, int64_t timestamp_intervals,
                        bool with_mvs) {
  int ret = DecoderBackend::Init(video_stream, codec, tgt_pix_fmt, tgt_w, tgt_h,
                                 sample_rate, timestamp_intervals, with_mvs);
  if (Status_OK != ret) return ret;

  // allocate AVCodecContext for AVCodec
  if (nullptr == (codec_ctx_ = avcodec_alloc_context3(codec))) {
    LOG(ERROR) << "allocate video codec failed";
    return Status_Error;
  }

  if (avcodec_parameters_to_context(codec_ctx_, video_stream->codecpar) < 0) {
    return Status_Error;
  }

  av_pix_fmt_ = FFPixelFormat(tgt_pix_fmt_);
  codec_ctx_->pix_fmt = av_pix_fmt_;
  int frame_size = codec_ctx_->width * codec_ctx_->height;
  if (frame_size <= 0) {
    LOG(WARNING) << "get frame width and height from codec_ctx failed";
  } else {
    codec_ctx_->thread_count = min(4, max(1, frame_size / (1920 * 1080)));
    LOG(INFO) << "decode  with " << codec_ctx_->thread_count << " threads";
  }

  AVDictionary *opts = NULL;
  if (with_mvs) av_dict_set(&opts, "flags2", "+export_mvs", 0);

  if (avcodec_open2(codec_ctx_, codec, &opts) < 0) {
    LOG(ERROR) << "couldn't open codec";
    av_dict_free(&opts);
    return Status_Error;
  }
  av_dict_free(&opts);
  src_w_ = codec_ctx_->width;
  src_h_ = codec_ctx_->height;

  av_frame_ = av_frame_alloc();
  if (av_frame_ == nullptr) {
    LOG(ERROR) << "allocate frame failed";
    return Status_Error;
  }

  return Status_OK;
}

void FFmpegBackend::Close() {
  DecoderBackend::Close();

  Free(codec_ctx_);
  Free(av_frame_);
  Free(decoded_frame_);

  if (nullptr != filter_graph_) {
    avfilter_graph_free(&filter_graph_);
    filter_graph_ = nullptr;
    buffersrc_ctx_ = nullptr;
    buffersink_ctx_ = nullptr;
  }
}

int FFmpegBackend::PutPacket(const AVPacket *packet, int64_t timestamp) {
  static char err_msg[50];
  if (0 == packet->size) return Status_OK;

  int ret = avcodec_send_packet(codec_ctx_, packet);
  if (ret < 0) {
    av_strerror(ret, err_msg, sizeof(err_msg));
    LOG(ERROR) << "Error while sending packet to decoder: " << err_msg;
    return Status_Error;
  }
  cur_timestamp_ = timestamp;
  return Status_OK;
}

int FFmpegBackend::Next(VFFrame &vf_frame) {
  static char err_msg[50];
  while (true) {
    int ret = avcodec_receive_frame(codec_ctx_, av_frame_);
    if (AVERROR(EAGAIN) == ret) {
      return Status_Pending;
    }
    if (AVERROR_EOF == ret) {
      return Status_EndOfFile;
    }
    if (ret < 0) {
      av_strerror(ret, err_msg, sizeof(err_msg));
      LOG(ERROR) << "Error while receiving a frame from the decoder: "
                 << err_msg;
      return Status_Error;
    }
    vf_frame.id = cur_frame_id_++;
    if (vf_frame.id % sample_rate_ != 0) continue;

    if (cur_timestamp() - vf_frame.timestamp < timestamp_interval_) {
      continue;
    } else {
      vf_frame.timestamp = cur_timestamp();
      break;
    }
  }

  return PostProcessFrame(vf_frame);
}

int FFmpegBackend::PostProcessFrame(VFFrame &vf_frame) {
  if (av_frame_->width <= 0 || av_frame_->height <= 0) {
    LOG(ERROR) << "PostProcessFrame (" << av_frame_->width << ","
               << av_frame_->height << ") failed";
    return Status_Error;
  }
  if (tgt_w_ < 0) tgt_w_ = av_frame_->width;
  if (tgt_h_ < 0) tgt_h_ = av_frame_->height;

  AVFrame *frame = av_frame_;
  if (av_frame_->width != tgt_w_ || av_frame_->height != tgt_h_ ||
      av_frame_->format != av_pix_fmt_) {
    // need post process
    if (nullptr == filter_graph_ || src_w_ != av_frame_->width ||
        src_h_ != av_frame_->height) {
      LOG(WARNING) << "re-create the filter graph";
      // need to re-create the filter graph
      if (nullptr != filter_graph_) avfilter_graph_free(&filter_graph_);
      src_w_ = av_frame_->width;
      src_h_ = av_frame_->height;
      if (Status_OK != InitFilters(av_frame_->format)) return Status_Error;
    }

    // push the decoded frame into the filtergraph
    if (av_buffersrc_add_frame_flags(buffersrc_ctx_, av_frame_,
                                     AV_BUFFERSRC_FLAG_KEEP_REF) < 0) {
      LOG(ERROR) << "Error while feeding the filtergraph";
      return Status_Error;
    }
    // pull filtered frames from the filtergraph
    av_frame_unref(decoded_frame_);
    while (true) {
      int ret = av_buffersink_get_frame(buffersink_ctx_, decoded_frame_);
      if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF) break;
      if (ret < 0) return Status_Error;
    }

    if (decoded_frame_->format != av_pix_fmt_) {
      LOG(ERROR) << "expect pix_fmt: " << decoded_frame_->format << ", but got "
                 << av_frame_->format;
      return Status_Error;
    }
    frame = decoded_frame_;
  }
  vf_frame.ctx.dev_type = kCPU;
  vf_frame.ctx.dev_id = -1;
  vf_frame.descr =
      PixelDescr<uint8_t>::Parse(vf_frame.descr.pix_fmt, frame->height,
                                 frame->width, frame->linesize, frame->data);

  vf_frame.pic_type = static_cast<int>(av_frame_->pict_type);
  AVFrameSideData *sd =
      av_frame_get_side_data(av_frame_, AV_FRAME_DATA_MOTION_VECTORS);
  if (nullptr != sd) {
    vf_frame.mvs = (uint8_t *)sd->data;
    vf_frame.mvs_length = sd->size;
  } else {
    vf_frame.mvs = nullptr;
    vf_frame.mvs_length = 0;
  }
  return Status_OK;
}

int FFmpegBackend::InitFilters(int src_pix_fmt) {
  char args[512];
  int ret = 0;
  ostringstream filter_descr_;
  const AVFilter *buffersrc = avfilter_get_by_name("buffer");
  const AVFilter *buffersink = avfilter_get_by_name("buffersink");
  filter_graph_ = avfilter_graph_alloc();
  AVFilterInOut *outputs = avfilter_inout_alloc();
  AVFilterInOut *inputs = avfilter_inout_alloc();
  enum AVPixelFormat pix_fmts[] = {av_pix_fmt_, AV_PIX_FMT_NONE};
  if (!outputs || !inputs || !filter_graph_) {
    ret = AVERROR(ENOMEM);
    goto end;
  }
  filter_graph_->nb_threads = 1;
  /* buffer video source: the decoded frames from the decoder will be inserted
   * here. */
  snprintf(args, sizeof(args),
           "video_size=%dx%d:pix_fmt=%d:time_base=%d/%d:pixel_aspect=%d/%d",
           src_w_, src_h_, src_pix_fmt, video_stream_->time_base.num,
           video_stream_->time_base.den, codec_ctx_->sample_aspect_ratio.num,
           codec_ctx_->sample_aspect_ratio.den);
  ret = avfilter_graph_create_filter(&buffersrc_ctx_, buffersrc, "in", args,
                                     NULL, filter_graph_);
  if (ret < 0) {
    LOG(ERROR) << "Cannot create buffer source";
    goto end;
  }

  /* buffer video sink: to terminate the filter chain. */
  ret = avfilter_graph_create_filter(&buffersink_ctx_, buffersink, "out", NULL,
                                     NULL, filter_graph_);
  if (ret < 0) {
    LOG(ERROR) << "Cannot create buffer sink";
    goto end;
  }
  ret = av_opt_set_int_list(buffersink_ctx_, "pix_fmts", pix_fmts,
                            AV_PIX_FMT_NONE, AV_OPT_SEARCH_CHILDREN);
  if (ret < 0) {
    LOG(ERROR) << "Cannot set output pixel formats";
    goto end;
  }

  /*
   * Set the endpoints for the filter graph. The filter_graph will
   * be linked to the graph described by filters_descr.
   */
  /*
   * The buffer source output must be connected to the input pad of
   * the first filter described by filters_descr; since the first
   * filter input label is not specified, it is set to "in" by
   * default.
   */
  outputs->name = av_strdup("in");
  outputs->filter_ctx = buffersrc_ctx_;
  outputs->pad_idx = 0;
  outputs->next = NULL;
  /*
   * The buffer sink input must be connected to the output pad of
   * the last filter described by filters_descr; since the last
   * filter output label is not specified, it is set to "out" by
   * default.
   */
  inputs->name = av_strdup("out");
  inputs->filter_ctx = buffersink_ctx_;
  inputs->pad_idx = 0;
  inputs->next = NULL;

  filter_descr_ << "scale=w=" << tgt_w_ << ":h=" << tgt_h_;

  if ((ret =
           avfilter_graph_parse_ptr(filter_graph_, filter_descr_.str().c_str(),
                                    &inputs, &outputs, NULL)) < 0) {
    LOG(ERROR) << "avfilter_graph_parse_ptr failed";
    goto end;
  }

  // connect the filters
  if ((ret = avfilter_graph_config(filter_graph_, NULL)) < 0) {
    LOG(ERROR) << "avfilter_graph_config failed";
    goto end;
  }
  decoded_frame_ = av_frame_alloc();
  if (decoded_frame_ == nullptr) {
    LOG(ERROR) << "allocate decoded frame failed";
    ret = -1;
    goto end;
  }

end:
  avfilter_inout_free(&inputs);
  avfilter_inout_free(&outputs);

  return ret < 0 ? Status_Error : Status_OK;
}

const string FFmpegBackend::kBackendName = "ffmpeg";

vector<BackendInfo> FFmpegBackend::GetBackendInfo() {
  BackendInfo c;
  c.dev_type = kCPU;
  c.dev_id = -1;
  c.limits = -1;
  c.priority = 0;
  c.name = FFmpegBackend::kBackendName;
  c.codecs = {-1};
  return {c};
}

RegisterDecoderBackend(FFmpegBackend, FFmpegBackend::kBackendName,
                       "ffmpeg decoder-backend, the most compatible one.");

}  // namespace io
}  // namespace vf

#endif
