/**
 * Copyright (c) 2020 xxx Inc.
 * File              : avio-decoder.cc
 * Author            : 
 * Date              : 2020-05-07
 * Last Modified Date: 2020-05-07
 * Last Modified By  : 
 */
#include "vf/io/avio-decoder.h"

#if USE_FFmpeg

#include <cxxutil/logging.h>
#include <cxxutil/memory-pool.h>
#include <cxxutil/strutil.h>

using namespace cxxutil;
using namespace std;

namespace vf {
namespace io {

int read_avio_packet(void *opaque, uint8_t *buf, int buf_size) {
  AVIODecoder *decoder = (AVIODecoder *)opaque;
  int read_size = 0;
  while (read_size < buf_size) {
    if (AVERROR_EOF == decoder->buf_size_) {
      if (0 == read_size)
        return AVERROR_EOF;
      else {
        return read_size;
      }
    } else if (0 == decoder->buf_size_) {
      uint8_t *buf = nullptr;
      int buf_len = decoder->read_packet_(decoder->istream_, &buf,
                                          &(decoder->cur_timestamp_));
      decoder->buf_size_ = buf_len;
      if (AVERROR_EOF == buf_len) return AVERROR_EOF;
      decoder->buf_ptr_ = buf;
    } else {
      int bs = FFMIN(buf_size - read_size, decoder->buf_size_);
      memcpy(buf + read_size, decoder->buf_ptr_, bs);
      decoder->buf_ptr_ += bs;
      decoder->buf_size_ -= bs;
      read_size += bs;
    }
  }
  return read_size;
}

int read_packet(void *opaque, uint8_t **buf, int64_t *timestamp) {
  AVIODecoder *decoder = (AVIODecoder *)opaque;
  if (nullptr != decoder->cur_stream_buf_) {
    MemoryPool<AVIODecoder::StreamBuf>::deallocate(decoder->cur_stream_buf_);
    decoder->cur_stream_buf_ = nullptr;
  }
  decoder->cur_stream_buf_ = decoder->stream_bufs_.Dequeue();

  if (nullptr == decoder->cur_stream_buf_) {
    return AVERROR_EOF;
  } else {
    *buf = &((decoder->cur_stream_buf_->buf)[0]);
    *timestamp = decoder->cur_stream_buf_->timestamp;
    return decoder->cur_stream_buf_->buf.size();
  }
}

AVIODecoder::AVIODecoder(ReadAVIOPacketHandler handler)
    : stream_bufs_(25, [](StreamBuf *&buf) {
        if (nullptr != buf) {
          MemoryPool<StreamBuf>::deallocate(buf);
          buf = nullptr;
        }
      }) {
  if (nullptr != handler) {
    read_avio_packet_ = handler;
  } else {
    read_avio_packet_ = read_avio_packet;
  }
}

AVIODecoder::~AVIODecoder() { Close(); }

bool AVIODecoder::Open(const std::string &uri, int pix_fmt, Context ctx,
                       int tgt_w, int tgt_h, int sample_rate,
                       int64_t timestamp_interval, bool with_mvs) {
  if (false == FFmpegDecoder::Open(uri, pix_fmt, ctx, tgt_w, tgt_h, sample_rate,
                                   timestamp_interval, with_mvs))
    return false;
  if (uri.empty()) {
    read_packet_ = read_packet;
    istream_ = (void *)(this);
    return true;
  }
  const vector<string> &paths = split(uri, ';');
  if (paths.size() != 2) {
    LOG(ERROR) << "path of AVIODecoder should be: "
                  "read://addr;istream://adddr, received: "
               << uri;
    return false;
  }
  static const vector<string> path_prefixes = {"read://", "istream://"};
  for (int i = 0; i < 2; ++i) {
    // check path format
    if (!startswith(paths[i], path_prefixes[i])) {
      LOG(ERROR) << "path of AVIODecoder should be: "
                    "read://addr;istream://adddr, received: "
                 << uri;
      return false;
    }
    try {
      long addr = stol(paths[i].substr(path_prefixes[i].length()));
      if (0 == i)
        read_packet_ = (ReadPacketHandler)(addr);
      else if (1 == i)
        istream_ = (void *)(addr);
    } catch (...) {
      LOG(ERROR) << "path of AVIODecoder should start with "
                    "\"itc://function_address\", but received: "
                 << uri;
    }
  }
  return true;
}

void AVIODecoder::Interrupt() {
  if (read_packet != read_packet_) {
    FFmpegDecoder::Interrupt();
  } else if (Status_UnInitialized == status_ || Status_OK == status_) {
    status_ = Status_EndOfFile;
    stream_bufs_.clear();
    stream_bufs_.Enqueue(nullptr);
  } else {
    LOG(ERROR) << "Interrupt is not allowed when status is " << status_;
  }
}

bool AVIODecoder::PutVideoPacket(const uint8_t *stream, int64_t stream_len,
                                 int64_t timestamp) {
  if (read_packet != read_packet_) {
    LOG(ERROR)
        << "PutVideoPacket is not allowed with user specified read_packet";
    return false;
  } else if (Status_UnInitialized == status_ || Status_OK == status_ ||
             true == is_reopening_) {
    if (stream_len < 0) {
      stream_bufs_.Enqueue(nullptr);
      return true;
    } else {
      StreamBuf *buf = MemoryPool<StreamBuf>::allocate();
      buf->buf.resize(stream_len);
      memcpy(&((buf->buf)[0]), stream, stream_len);
      buf->timestamp = timestamp;
      stream_bufs_.Enqueue(buf);
      return true;
    }
  }
  return false;
}

bool AVIODecoder::OpenFormatContext() {
  avio_ctx_buffer_ = (uint8_t *)av_malloc(avio_ctx_buffer_size_);
  if (avio_ctx_buffer_ == nullptr) {
    LOG(ERROR) << "av_malloc  avio_ctx_buffer_ failed";
    Close();
    return false;
  }

  avio_ctx_ = avio_alloc_context(avio_ctx_buffer_, avio_ctx_buffer_size_, 0,
                                 this, read_avio_packet_, NULL, NULL);
  if (avio_ctx_ == nullptr) {
    LOG(ERROR) << "avio_alloc_context failed";
    Close();
    return false;
  }
  format_ctx_->pb = avio_ctx_;
  format_ctx_->flags = AVFMT_FLAG_CUSTOM_IO;

  static char err_msg[50];
  int ret;
  if ((ret = avformat_open_input(&format_ctx_, nullptr, nullptr, nullptr)) <
      0) {
    av_strerror(ret, err_msg, sizeof(err_msg));
    LOG(ERROR) << "open video input failed: " << err_msg;
    Close();
    return false;
  }
  return true;
}

void AVIODecoder::Close() {
  status_ = Status_EndOfFile;
  stream_bufs_.clear();
  if (nullptr != cur_stream_buf_) {
    MemoryPool<StreamBuf>::deallocate(cur_stream_buf_);
    cur_stream_buf_ = nullptr;
  }

  FFmpegDecoder::Close();
  if (avio_ctx_ != nullptr) {
    av_freep(&avio_ctx_->buffer);
    av_freep(&avio_ctx_);
  }
}

RegisterVideoDecoder(AVIODecoder, "avio", "video decoder from memory stream");

}  // namespace io
}  // namespace vf

#endif
