/**
 * Copyright (c) 2020 xxx Inc.
 * File              : http-decoder.cc
 * Author            : 
 * Date              : 2020-05-07
 * Last Modified Date: 2020-05-07
 * Last Modified By  : 
 */
#include "vf/io/avio-decoder.h"

//
#include <curl/curl.h>
#include <cxxutil/logging.h>
#include <cxxutil/types.h>

using namespace std;
using namespace cxxutil;

namespace vf {
namespace io {

size_t WriteMemoryCallback(void *contents, size_t size, size_t nmemb,
                           void *userp);

class HTTPDecoder : public AVIODecoder {
 public:
  virtual ~HTTPDecoder() override { Close(); }

  bool Open(const std::string &uri, int pix_fmt = VF_PIX_FMT_BGR,
            Context ctx = {kAuto, -1}, int tgt_w = -1, int tgt_h = -1,
            int sample_rate = 1, int64_t timestamp_interval = 0,
            bool with_mvs = false) override;

  VFFrame *Next() override;

  friend size_t WriteMemoryCallback(void *contents, size_t size, size_t nmemb,
                                    void *userp);

 protected:
  void Close() override;

 private:
  VFFrame *first_frame_ = nullptr;
  CURL *curl_handle_ = nullptr;
  shared_ptr<std::thread> download_task_;
};

bool HTTPDecoder::Open(const std::string &uri, int pix_fmt, Context ctx,
                       int tgt_w, int tgt_h, int sample_rate,
                       int64_t timestamp_interval, bool with_mvs) {
  if (false == AVIODecoder::Open("", pix_fmt, ctx, tgt_w, tgt_h, sample_rate,
                                 timestamp_interval, with_mvs)) {
    return false;
  }

  /* init the curl session */
  curl_handle_ = curl_easy_init();

  /* specify URL to get */
  curl_easy_setopt(curl_handle_, CURLOPT_URL, uri.c_str());
  /* complete connection within 10 seconds */
  curl_easy_setopt(curl_handle_, CURLOPT_CONNECTTIMEOUT, 10L);
  /* send all data to this function  */
  curl_easy_setopt(curl_handle_, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);
  /* we pass our 'chunk' struct to the callback function */
  curl_easy_setopt(curl_handle_, CURLOPT_WRITEDATA, this);
  /* some servers don't like requests that are made without a user-agent
     field, so we provide one */
  curl_easy_setopt(curl_handle_, CURLOPT_USERAGENT, "libcurl-agent/1.0");

  uri_ = uri;

  download_task_.reset(new std::thread(
      [uri, this](void *) {
        /* get it! */
        CURLcode res = curl_easy_perform(curl_handle_);
        if (CURLE_OK != res) {
          LOG(ERROR) << "open url " << uri
                     << " failed: " << curl_easy_strerror(res);
        }
        PutVideoPacket(nullptr, -1, 0);
      },
      nullptr));

  first_frame_ = Next();
  return nullptr == first_frame_ ? false : true;
}

VFFrame *HTTPDecoder::Next() {
  if (nullptr == first_frame_) {
    return AVIODecoder::Next();
  } else {
    VFFrame *ret = first_frame_;
    first_frame_ = nullptr;
    return ret;
  }
}

void HTTPDecoder::Close() {
  AVIODecoder::Close();
  if (download_task_) {
    download_task_->join();
    download_task_.reset();
  }
  /* cleanup curl stuff */
  if (nullptr != curl_handle_) {
    curl_easy_cleanup(curl_handle_);
    curl_handle_ = nullptr;
  }
}
size_t WriteMemoryCallback(void *contents, size_t size, size_t nmemb,
                           void *userp) {
  HTTPDecoder *decoder = (HTTPDecoder *)userp;
  if (!(Status_UnInitialized == decoder->status_ ||
        Status_OK == decoder->status_)) {
    return CURL_READFUNC_ABORT;
  } else {
    decoder->PutVideoPacket((uint8_t *)contents, size * nmemb, 0);
    return size * nmemb;
  }
}

RegisterVideoDecoder(HTTPDecoder, "http", "read video stream from http");

}  // namespace io
}  // namespace vf
