/**
 * Copyright (c) 2020 xxx Inc.
 * File              : image-decoder.cc
 * Author            : 
 * Date              : 2020-05-07
 * Last Modified Date: 2020-05-07
 * Last Modified By  : 
 */
#include "vf/flow/image-decoder.h"

#include <cxxutil/logging.h>
#include <cxxutil/memory-pool.h>

#include <opencv2/opencv.hpp>

#include "vf/vf-frame.h"

#if CV_VERSION_MAJOR > 3
#define CV_LOAD_IMAGE_COLOR cv::IMREAD_COLOR
#endif

using namespace cxxutil;
using namespace std;

namespace vf {
namespace flow {

ImageDecoder::ImageDecoder(const vector<pair<string, string>> &params)
    : Entry({"VFFrame"}) {
  param_.Init(params, parameter::kAllowUnknown);
}

ImageDecoder::~ImageDecoder() { DeletePointer(tasks_); }

int ImageDecoder::Init(int max_orders) {
  int ret = Entry::Init(max_orders);
  if (Status_OK == ret) {
    tasks_ = new BlockQueue<ImageTask *>(max_orders + 1, [](ImageTask *&task) {
      if (nullptr != task) {
        task->OnFinish();
      }
    });
  }
  return ret;
}

void ImageDecoder::Interrupt() {
  tasks_->clear();
  tasks_->Enqueue(nullptr);
  Entry::Interrupt();
}

shared_ptr<ImageTask> ImageDecoder::PutImage(const uint8_t *img_stream,
                                             int64_t stream_len,
                                             const std::string &channel_id,
                                             int64_t timestamp) {
  if (Status_OK != status_.load()) {
    LOG(INFO) << "return nullptr";
    return nullptr;
  }

  ImageTask *task = MemoryPool<ImageTask>::allocate();
  task->Reset();
  std::vector<char> img_buf((const char *)img_stream,
                            (const char *)img_stream + stream_len);
  int status = Status_OK;
  try {
    task->img = cv::imdecode(img_buf, CV_LOAD_IMAGE_COLOR);
  } catch (cv::Exception &ex) {
    LOG(INFO) << "imdecode failed: " << ex.what();
    status = Status_IOError;
  }
  if (Status_OK != status || nullptr == task->img.data) {
    LOG(ERROR) << "decode image failed";
    MemoryPool<ImageTask>::deallocate(task);
    return nullptr;
  }
  task->frame = task->img;
  task->frame.timestamp = timestamp;
  task->frame.video_id = channel_id;
  task->frame.uuid = VFFrame::UUID(task->frame);
  task->frame.extra = task;

  tasks_->Enqueue(task);

  return shared_ptr<ImageTask>(
      task, [](ImageTask *t) { MemoryPool<ImageTask>::deallocate(t); });
}

vector<void *> ImageDecoder::run(const DataType &res_data) {
  ImageTask *task = tasks_->Dequeue();
  if (nullptr == task) {
    NotifyEOF();
    return {nullptr};
  } else {
    return {&(task->frame)};
  }
}

void ImageDecoder::ReleaseOutput(vector<void *> &output) {
  ImageTask *task = (ImageTask *)(output[0]);
  if (nullptr != task) task->OnFinish();
}

CXXUTIL_REGISTER_PARAMETER(ImageDecoderParam);
RegisterOp(ImageDecoder, "ImageDecoder", "Image decoder",
           ImageDecoderParam::GetInfo(), "VF", "entry");

}  // namespace flow
}  // namespace vf
