/**
 * Copyright (c) 2020 xxx Inc.
 * File              : video-decoder.cc
 * Author            : 
 * Date              : 2020-05-07
 * Last Modified Date: 2020-05-07
 * Last Modified By  : 
 */
#include "vf/flow/video-decoder.h"

#include <cxxutil/logging.h>

#include "vf/flow/engine.h"

using namespace cxxutil;
using namespace std;

namespace vf {
namespace flow {

VideoDecoder::VideoDecoder(const vector<pair<string, string>> &params)
    : StreamTaskEntry({"VFFrame"}) {
  param_.Init(params);
}

VideoDecoder::~VideoDecoder() {
  DeletePointer(decoder_);
  if (kGPU == ctx_.dev_type && nullptr != stream_) {
    cuda::DestroyStream(ctx_.dev_id, stream_);
    stream_ = nullptr;
  }
}

int VideoDecoder::Open(const vector<string> &params) {
  if (params.size() != 2) {
    LOG(ERROR) << "VideoDeocder requires two params to open, one is uri, the "
                  "other is decoder type, but got "
               << params;
    return Status_InvalidArgument;
  }
  const string &uri = params[0];
  const string &type = params[1];
  try {
    decoder_ = io::VideoDecoder::Create(type);
  } catch (exception &err) {
    LOG(ERROR) << err.what();
    return Status_InvalidArgument;
  }

  if (nullptr == decoder_ ||
      false == decoder_->Open(uri, param_.pixel_format, {param_.decoder, -1},
                              param_.width, param_.height, param_.sample_rate,
                              param_.timestamp_interval, param_.with_mvs)) {
    DeletePointer(decoder_);
    return Status_InvalidArgument;
  }
  first_frame_ = true;
  decoder_->set_video_id(channel_);
  should_stop_.store(false);
  return Status_OK;
}

void VideoDecoder::Interrupt() {
  should_stop_.store(true);
  if (nullptr != decoder_) decoder_->Interrupt();
  StreamTaskEntry::Interrupt();
}

int VideoDecoder::PutData(int64_t timestamp, uint8_t *data_stream,
                          int64_t stream_len, void *extra) {
  if (should_stop_.load() || status_.load() != Status_OK) return Status_Error;

  LockGuard guard(monitor_);
  if (decoder_->PutVideoPacket(data_stream, stream_len, timestamp)) {
    return Status_OK;
  } else {
    should_stop_.store(true);
    return Status_Error;
  }
}

Op::DataType VideoDecoder::run(const Op::DataType &res_data) {
  VFFrame &dst_frame = Get<VFFrame>(res_data, 0);

  VFFrame *frame = decoder_->Next();
  if (nullptr == frame) {
    int decoder_status = decoder_->status();
    if (Status_OK != decoder_status && Status_EOF != decoder_status) {
      OnError(decoder_status, "video decoder failed");
    } else {
      NotifyEOF();
    }
    return {nullptr};
  }

  if (first_frame_) {
    first_frame_ = false;
    if (frame->ctx.dev_type == kGPU) {
      stream_ = cuda::CreateStream(frame->ctx.dev_id);
    }
    for (int i = 0; i < max_orders_; ++i) {
      Get<VFFrame>(resources_[i], 0).stream = stream_;
    }

    ctx_ = frame->ctx;
  }
  frame->CopyTo(&dst_frame);
  return {&dst_frame};
}

vector<void *> VideoDecoder::MallocResource() {
  VFFrame *frame = new VFFrame;
  frame->video_id = channel_;
  return {frame};
}

void VideoDecoder::FreeResource(vector<void *> output) {
  CHECK(output.size() == 1)
      << "runtime error, output should be size 1, but got " << output.size();
  VFFrame *frame = (VFFrame *)(output[0]);
  delete frame;
  output.clear();
}

CXXUTIL_REGISTER_PARAMETER(VideoDecodeParam);
RegisterOp(VideoDecoder, "VideoDecoder", "Video decoder",
           VideoDecodeParam::GetInfo(), "VF", "entry");

}  // namespace flow
}  // namespace vf
