/*
 * Author:  
 * Created Date: 2018-10-23
 * Modified By: 
 * Last Modified: 2019-06-18
 * -----
 * Copyright (c) 2018 xxx Inc.
 */

#include <cxxutil/logging.h>

#include "vf/dl/dl-processor.h"
#include "vf/dl/types.h"
#include "vf/flow/op.h"
#include "vf/resource/resource.h"
#include "vf/vf-frame.h"

using namespace cxxutil;
using namespace std;
using namespace vf::dl;

namespace vf {
namespace flow {

struct DetectorParam : public Parameter<DetectorParam> {
  int det_interv;
  bool abs_coords;
  int min_size;
  ResourceType resource;
  CXXUTIL_DECLARE_PARAMETER(DetectorParam) {
    CXXUTIL_DECLARE_FIELD(det_interv)
        .set_default(1)
        .describe("detection interval");
    CXXUTIL_DECLARE_FIELD(abs_coords)
        .set_default(false)
        .describe(
            "if true, absolute coordinate will be considered, otherwise "
            "relative coordinate");
    CXXUTIL_DECLARE_FIELD(min_size)
        .set_lower_bound(10)
        .set_default(10)
        .describe("minimum size of the detected boxes");
    CXXUTIL_DECLARE_FIELD(resource).describe(
        "resource name (name of the DLProcessor)");
  }
};

class Detector : public Op {
 public:
  Detector(const vector<pair<string, string>> &params = {},
           const vector<Op *> &ops = {})
      : Op(false, {{"VFFrame"}}, {"VFFrame", "vector<DLObject>"}, 1, ops) {
    param_.Init(params);
  }

  virtual ~Detector() {}

  int Init(int max_orders) override {
    // request DLProcessor
    Resource *res = ResourceManager::Request(param_.resource);
    if (nullptr == res) {
      LOG(ERROR) << "Request resource " << param_.resource << " failed";
      return Status_InvalidArgument;
    }
    processor_ = res->cast<DLProcessor>();
    return Op::Init(max_orders);
  }

 protected:
  vector<void *> run(const vector<vector<void *>> &in_data,
                     const vector<void *> &res_data) override {
    VFFrame &frame = Get<VFFrame>(in_data[0], 0);
    vector<DLObject> &objs = Get<vector<DLObject>>(res_data, 0);
    objs.clear();

    if (frame.id % param_.det_interv != 0) {
      return {&frame, &objs};
    }

    DLTask *task = processor_->Acquire(frame.ctx);
    if (nullptr == task) {
      LOG(ERROR) << "Preprocess image failed";
      return {&frame, &objs};
    }
    frame.ConvertTo(task, 0);

    processor_->Process(task);
    float w_scale = frame.descr.w, h_scale = frame.descr.h;
    if (param_.abs_coords) {
      w_scale /= task->shapes[0][2];
      h_scale /= task->shapes[0][1];
    }

    // 6 dims: [label, score, xmin, ymin, xmax, ymax]
    size_t box_num = task->result.size() / 6;
    objs.resize(box_num);
    const float *src_box = task->result.data();
    int valid_obj_num = 0;
    for (size_t i = 0; i < box_num; ++i, src_box += 6) {
      if (src_box[0] < 0) continue;
      DLObject &dst_obj = objs[valid_obj_num];
      dst_obj.Reset();
      dst_obj.pos.xmin = int(w_scale * src_box[2]);
      dst_obj.pos.ymin = int(h_scale * src_box[3]);
      dst_obj.pos.xmax = int(w_scale * src_box[4]);
      dst_obj.pos.ymax = int(h_scale * src_box[5]);
      if (dst_obj.pos.xmin < 0 || dst_obj.pos.ymin < 0 ||
          dst_obj.pos.xmax >= frame.descr.w ||
          dst_obj.pos.ymax >= frame.descr.h ||
          dst_obj.pos.xmax - dst_obj.pos.xmin < param_.min_size ||
          dst_obj.pos.ymax - dst_obj.pos.ymin < param_.min_size) {
        continue;
      }
      dst_obj.type = src_box[0];
      dst_obj.score = src_box[1];
      dst_obj.timestamp = frame.timestamp;
      dst_obj.entry_time = frame.timestamp;
      dst_obj.camera_id = frame.video_id;
      dst_obj.id =
          frame.uuid + "_" + to_string(i) + "_" + to_string(rand() % 65536);

      valid_obj_num++;
    }
    objs.resize(valid_obj_num);
    processor_->Return(task);
    return {&frame, &objs};
  }

  vector<void *> MallocResource() override { return {new vector<DLObject>}; }
  void FreeResource(vector<void *> output) override {
    CHECK(output.size() == 1)
        << "invalid call to FreeResource, expected size 1, but got "
        << output.size();
    for (void *ptr : output) {
      delete (vector<DLObject> *)ptr;
    }
  }

 protected:
  DetectorParam param_;
  DLProcessor *processor_ = nullptr;
};

CXXUTIL_REGISTER_PARAMETER(DetectorParam);

RegisterOp(Detector, "Detector", "Detect Object", DetectorParam::GetInfo(),
           "VF", "op");

}  // namespace flow
}  // namespace vf
