/**
 * Copyright (c) 2020 xxx Inc.
 * File              : yuv.h
 * Author            : 
 * Date              : 2020-05-07
 * Last Modified Date: 2020-05-07
 * Last Modified By  : 
 */
#include "vf/pixcvt.h"

namespace vf {

XINLINE float f_min(float lhs, float rhs) { return lhs < rhs ? lhs : rhs; }
XINLINE float f_max(float lhs, float rhs) { return lhs > rhs ? lhs : rhs; }

XINLINE bool isbgr(int pixfmt) {
  return pixfmt == VF_PIX_FMT_BGR || pixfmt == VF_PIX_FMT_BGRPlanar;
}

XINLINE void YUV2RGB(uint8_t *yuv, uint8_t *rgb, bool swap) {
  const float luma = float(yuv[0]);
  const float u = float(yuv[1]) - 128.0f;
  const float v = float(yuv[2]) - 128.0f;

  // BT.709, for YCbCr
  if (swap) {
    rgb[2] = uint8_t(
        f_min(f_max(0, 1.164 * (luma - 16) + 1.792f * v), 255.f));  // red
    rgb[1] =
        uint8_t(f_min(f_max(0, 1.164 * (luma - 16) - 0.213f * u - 0.534f * v),
                      255.f));  // greeen
    rgb[0] = uint8_t(
        f_min(f_max(0, 1.164 * (luma - 16) + 2.114f * u), 255.f));  // blue
  } else {
    rgb[0] = uint8_t(
        f_min(f_max(0, 1.164 * (luma - 16) + 1.792f * v), 255.f));  // red
    rgb[1] =
        uint8_t(f_min(f_max(0, 1.164 * (luma - 16) - 0.213f * u - 0.534f * v),
                      255.f));  // greeen
    rgb[2] = uint8_t(
        f_min(f_max(0, 1.164 * (luma - 16) + 2.114f * u), 255.f));  // blue
  }

  /* for YUV
  if (swap) {
    rgb[2] = uint8_t(f_min(f_max(0, luma + 1.140f * v), 255.f));  // red
    rgb[1] = uint8_t(
        f_min(f_max(0, luma - 0.395f * u - 0.581f * v), 255.f));  // greeen
    rgb[0] = uint8_t(f_min(f_max(0, luma + 2.032f * u), 255.f));  // blue
  } else {
    rgb[0] = uint8_t(f_min(f_max(0, luma + 1.140f * v), 255.f));  // red
    rgb[1] = uint8_t(
        f_min(f_max(0, luma - 0.395f * u - 0.581f * v), 255.f));  // greeen
    rgb[2] = uint8_t(f_min(f_max(0, luma + 2.032f * u), 255.f));  // blue
  }
  */
}

/**
 * \brief  Get Cb Cr values
 *
 * \tparam T data formats
 */
template <int PixFmt>
struct YCbCr;

template <>
struct YCbCr<VF_PIX_FMT_NV12> {
  /**
   * \brief  Get Cb Cr values from NV12
   *
   * \param src source NV12 pixel descriptor
   * \param x x coordinates, must be even
   * \param y y coordinates
   * \param Cb Cb value
   * \param Cr Cr value
   *
   * \return none
   */
  static XINLINE void Get(const PixelDescr<uint8_t> *src, int x, int y,
                          uint8_t *Cb, uint8_t *Cr) {
    int y_chroma = y >> 1;
    int x_chroma = (x >> 1) << 1;
    *Cb = (src->data[1])[y_chroma * src->step[1] + x_chroma];
    *Cr = (src->data[1])[y_chroma * src->step[1] + x_chroma + 1];
    ++y_chroma;
    if ((y & 1) && (y_chroma < ((src->h >> 1)))) {
      // odd scanline, interpolate chroma vertically
      *Cb = (*Cb + (src->data[1])[y_chroma * src->step[1] + x_chroma] + 1) >> 1;
      *Cr =
          (*Cr + (src->data[1])[y_chroma * src->step[1] + x_chroma + 1] + 1) >>
          1;
    }
  }
};

template <>
struct YCbCr<VF_PIX_FMT_NV21> {
  /**
   * \brief  Get Cb Cr values from NV21
   *
   * \param src source NV21 pixel descriptor
   * \param x x coordinates, must be even
   * \param y y coordinates
   * \param Cb Cb value
   * \param Cr Cr value
   *
   * \return none
   */
  static XINLINE void Get(const PixelDescr<uint8_t> *src, int x, int y,
                          uint8_t *Cb, uint8_t *Cr) {
    int y_chroma = y >> 1;
    int x_chroma = (x >> 1) << 1;
    *Cr = (src->data[1])[y_chroma * src->step[1] + x_chroma];
    *Cb = (src->data[1])[y_chroma * src->step[1] + x_chroma + 1];
    ++y_chroma;
    if ((y & 1) && (y_chroma < ((src->h >> 1)))) {
      // odd scanline, interpolate chroma vertically
      *Cr = (*Cr + (src->data[1])[y_chroma * src->step[1] + x_chroma] + 1) >> 1;
      *Cb =
          (*Cb + (src->data[1])[y_chroma * src->step[1] + x_chroma + 1] + 1) >>
          1;
    }
  }
};

template <>
struct YCbCr<VF_PIX_FMT_YUV420P> {
  /**
   * \brief  Get Cb Cr values from YUV420P
   *
   * \param src source YUV420P pixel descriptor
   * \param x x coordinates, must be even
   * \param y y coordinates
   * \param Cb Cb value
   * \param Cr Cr value
   *
   * \return none
   */
  static XINLINE void Get(const PixelDescr<uint8_t> *src, int x, int y,
                          uint8_t *Cb, uint8_t *Cr) {
    int x_chroma = x >> 1;
    int y_chroma = y >> 1;
    *Cb = (src->data[1])[y_chroma * src->step[1] + x_chroma];
    *Cr = (src->data[2])[y_chroma * src->step[2] + x_chroma];
    ++y_chroma;
    if ((y & 1) && (y_chroma < ((src->h >> 1)))) {
      // odd scanline, interpolate chroma vertically
      *Cb = (*Cb + (src->data[1])[y_chroma * src->step[1] + x_chroma] + 1) >> 1;
      *Cr = (*Cr + (src->data[2])[y_chroma * src->step[2] + x_chroma] + 1) >> 1;
    }
  }
};

template <int sPixFmt>
XINLINE void GetRGB(const PixelDescr<uint8_t> *src, int x, int y, uint8_t *rgb,
                    bool swap) {
  uint8_t yuv[3];

  // Read 2 Luma components at a time, so we don't waste processing since CbCr
  // are decimated this way.
  // if we move to texture we could read 4 luminance values
  yuv[0] = ((src->data[0])[y * src->step[0] + x]);
  YCbCr<sPixFmt>::Get(src, x, y, yuv + 1, yuv + 2);

  YUV2RGB(yuv, rgb, swap);

  yuv[0] = ((src->data[0])[y * src->step[0] + x + 1]);
  if ((x & 1)) {
    // if x is odd, we need to re-get the Cb Cr values
    YCbCr<sPixFmt>::Get(src, x + 1, y, yuv + 1, yuv + 2);
  }
  YUV2RGB(yuv, rgb + 3, swap);
}

template <typename T, int sPixFmt>
XINLINE void YUV2RGBConvFunc(int dst_idx, int dx, int dy,
                             const PixelDescr<uint8_t> *src,
                             const CoordMapParam *mparam,
                             PixelDescr<T> const *dst, bool swap) {
  // coordinates in the source image
  float sxf = dx * mparam->fx + mparam->sx;
  float syf = dy * mparam->fy + mparam->sy;
  int sx = int(sxf), sy = int(syf);

  // interpolation coefficients
  float alpha = sxf - sx, beta = syf - sy;

  // process the x-line
  uint8_t rgb[12];  // 2 * 3
  if (sx + 1 < src->w) {
    GetRGB<sPixFmt>(src, sx, sy, rgb, swap);
  } else {
    GetRGB<sPixFmt>(src, sx - 1, sy, rgb, swap);
  }
  if (sy + 1 < src->h) {
    GetRGB<sPixFmt>(src, sx, sy + 1, rgb + 6, swap);
  } else {
    for (int i = 0; i < 6; ++i) {
      rgb[i + 6] = rgb[i];
    }
  }

  for (int c = 0; c < 3; ++c) {
    T *dptr = (T *)(dst->data[c]) + dy * dst->step[c] + int(dx * dst->pstep[c]);

    float val1 = (1 - alpha) * rgb[c] + alpha * rgb[c + 3];
    float val2 = (1 - alpha) * rgb[c + 6] + alpha * rgb[c + 9];
    *dptr = T((1 - beta) * val1 + beta * val2);
  }
  for (int c = 3; c < dst->channels; ++c) {
    T *dptr = (T *)(dst->data[c]) + dy * dst->step[c] + int(dx * dst->pstep[c]);
    *dptr = 0;
  }
}

#define RegisterYUVCvtFunc(KDevice, sPixFmt, dPixFmt)                \
  RegisterPixelCvtFunc(KDevice, uint8_t, uint8_t, sPixFmt, dPixFmt); \
  RegisterPixelCvtFunc(KDevice, uint8_t, float, sPixFmt, dPixFmt);

}  // namespace vf
