/**
 * Copyright (c) 2020 xxx Inc.
 * File              : rocketmq-connector.cc
 * Author            : 
 * Date              : 2020-05-07
 * Last Modified Date: 2020-05-07
 * Last Modified By  : 
 */
#include "vf/resource/rocketmq-connector.h"

#ifdef USE_RocketMQ

#include <cxxutil/logging.h>
#include <cxxutil/timer.h>

using namespace std;
using namespace cxxutil;
using namespace rocketmq;

namespace vf {
namespace flow {

class SendMessageTask : public ThreadTask {
 public:
  SendMessageTask(RocketMQConnector *parent)
      : parent_(parent), producer_(parent->producer_.get()) {}

 protected:
  void run() override;
  void Send();

 protected:
  RocketMQConnector *parent_ = nullptr;
  DefaultMQProducer *producer_ = nullptr;
};

void SendMessageTask::run() {
  LockGuard guard(parent_->timer_);
  while (false == IsStop()) {
    if (parent_->cur_msg_num_ > 0) {
      Send();
      parent_->cur_msg_num_ = 0;
      parent_->msgs_.resize(parent_->param_.buf_size);
      parent_->timer_.notify_all();
    }
    parent_->timer_.wait_until(Timer::Now() + parent_->param_.max_latency);
  }

  if (parent_->cur_msg_num_ > 0) Send();
  parent_->cur_msg_num_ = 0;
}

void SendMessageTask::Send() {
  try {
    int64_t tic = Timer::Now();
    for (int i = 0; i < parent_->cur_msg_num_; ++i) {
      SendResult send_result = producer_->send(parent_->msgs_[i]);
      if (send_result.getSendStatus() != SEND_OK) {
        LOG(ERROR) << "send message failed with error code: "
                   << send_result.getSendStatus();
      }
    }
    int64_t duration = Timer::Now() - tic;
    if (duration >= 500) {
      LOG(WARNING) << "send " << parent_->cur_msg_num_
                   << " messages to rocketMQ for more than: " << duration
                   << " ms.";
      //} else {
      //  LOG(INFO) << "send " << parent_->cur_msg_num_ << " messages to
      //  rocketMQ";
    }
  } catch (const MQException &e) {
    LOG(ERROR) << "send failed: " << e.what();
  }
}

RocketMQConnector::~RocketMQConnector() {
  if (nullptr != send_task_) {
    send_task_->Stop();
    send_task_->Join();
    delete send_task_;
  }
  if (producer_) {
    producer_->shutdown();
  }
}

int RocketMQConnector::Init(const string &name, const ParamType &params) {
  int ret = Resource::Init(name, params);
  if (Status_OK != ret) return ret;
  param_.Init(params);

  producer_.reset(new DefaultMQProducer("videoflow"));
  producer_->setNamesrvAddr(param_.namesrv);
  producer_->setNamesrvDomain("");
  producer_->setGroupName(param_.group);
  producer_->setInstanceName(param_.group);
  if (false == param_.access_key.empty()) {
    producer_->setSessionCredentials(param_.access_key, param_.sceret_key,
                                     param_.ons_channel);
  }
  producer_->setSendMsgTimeout(500);
  producer_->setTcpTransportTryLockTimeout(1000);
  producer_->setTcpTransportConnectTimeout(400);
  producer_->start();
  send_task_ = new SendMessageTask(this);
  send_task_->Start();
  msgs_.resize(param_.buf_size);
  cur_msg_num_ = 0;
  return Status_OK;
}

int RocketMQConnector::WriteMessage(const string &topic, const string &msg) {
  LockGuard guard(timer_);
  while (cur_msg_num_ >= param_.buf_size) {
    timer_.notify_all();
    timer_.wait();
  }
  msgs_[cur_msg_num_].setTopic(topic);
  msgs_[cur_msg_num_].setBody(msg);
  cur_msg_num_++;
  return Status_OK;
}

CXXUTIL_REGISTER_PARAMETER(RocketMQConnectorParam);
RegisterResource(RocketMQConnector, "RocketMQConnector", "rocket mq connector",
                 RocketMQConnectorParam::GetInfo());
}  // namespace flow
}  // namespace vf
#endif
