/**
 * Copyright (c) 2020 xxx Inc.
 * File              : local-image-sinker.cc
 * Author            : 
 * Date              : 2020-05-07
 * Last Modified Date: 2020-05-07
 * Last Modified By  : 
 */
#include <sstream>

//
#include <opencv2/opencv.hpp>

//
#include <cxxutil/logging.h>
#include <cxxutil/os-path.h>
#include <cxxutil/timer.h>

#include "vf/dl/types.h"
#include "vf/flow/op.h"
#include "vf/vf-frame.h"

#if CV_VERSION_MAJOR > 3
#define CV_IMWRITE_JPEG_QUALITY cv::IMWRITE_JPEG_QUALITY
#endif

using namespace cxxutil;
using namespace std;
using namespace vf::dl;

namespace vf {
namespace flow {

struct LocalImageSinkerParam : public Parameter<LocalImageSinkerParam> {
  string root_dir;
  int jpeg_quality;
  string ori_img_dir;
  string cropped_img_dir;
  vector<int> compression_params;
  CXXUTIL_DECLARE_PARAMETER(LocalImageSinkerParam) {
    CXXUTIL_DECLARE_FIELD(root_dir).describe(
        "root directory to save the images");
    CXXUTIL_DECLARE_FIELD(jpeg_quality)
        .set_default(85)
        .describe("jpeg quality for compression");
  }
};

class LocalImageSinker : public Op {
 public:
  LocalImageSinker(const vector<pair<string, string>> &params = {})
      : Op(false, {{"VFFrame", "vector<DLObject>"}}, {}, 1) {
    param_.Init(params);
    param_.ori_img_dir = join_path(param_.root_dir, "orginal/");
    param_.cropped_img_dir = join_path(param_.root_dir, "cropped/");
    if (false == mkdirs(param_.ori_img_dir) ||
        false == mkdirs(param_.cropped_img_dir)) {
      ostringstream oss;
      oss << "create subdirectories in " << param_.root_dir << " failed";
      throw runtime_error(oss.str());
    }
    param_.compression_params = {CV_IMWRITE_JPEG_QUALITY, param_.jpeg_quality};
  }

  virtual ~LocalImageSinker() {}

 protected:
  vector<void *> run(const vector<vector<void *>> &in_data,
                     const vector<void *> &res_data) override {
    VFFrame &frame = Get<VFFrame>(in_data[0], 0);
    vector<DLObject> &objs = Get<vector<DLObject>>(in_data[0], 1);

    int best_num = 0;
    for (const DLObject &obj : objs) {
      if (false == obj.is_best) continue;
      ++best_num;
    }
    if (best_num > 0) {
      CHECK(false == frame.uuid.empty()) << "frame id must not be null";
      const string &ori_path = param_.ori_img_dir + frame.uuid + ".jpg";
      if (false == mkdirs(dirname(ori_path))) {
        LOG(ERROR) << "mkdirs for " << ori_path << " failed";
        return {};
      }

      const cv::Mat &img = frame.cv_img();
      if (!cv::imwrite(ori_path, img, param_.compression_params)) {
        LOG(ERROR) << "Save original image to " << ori_path << " failed";
        return {};
      }

      for (const DLObject &obj : objs) {
        if (false == obj.is_best) continue;
        CHECK(false == obj.id.empty()) << "object id must not be null";
        const string &crop_path = param_.cropped_img_dir + obj.id + ".jpg";
        if (false == mkdirs(dirname(crop_path))) {
          LOG(ERROR) << "mkdirs for " << crop_path << " failed";
          return {};
        }

        if (!cv::imwrite(crop_path, img(BBox2Rect(obj.pos)),
                         param_.compression_params)) {
          LOG(ERROR) << "Save object image to " << crop_path << " failed";
          return {};
        }
      }
    }

    return {};
  }

  LocalImageSinkerParam param_;
};

CXXUTIL_REGISTER_PARAMETER(LocalImageSinkerParam);
RegisterOp(LocalImageSinker, "LocalImageSinker", "Save local images",
           LocalImageSinkerParam::GetInfo(), "VF", "sinker");

}  // namespace flow
}  // namespace vf
